/* THIS FILE HAS BEEN AUTOMATICALLY GENERATED BY LLnextgen. DO NOT EDIT */
#ifndef  __LLNEXTGEN_LL_H__
#define __LLNEXTGEN_LL_H__
#ifndef LL_NOTOKENS
#define EOFILE 256
#define ENV 257
#define START 258
#define WINDOW_SIZE 259
#define EXPECT 260
#define SEND 261
#define EXPECT_EXIT 262
#define INTERACT 263
#define NUMBER 264
#define STRING 265
#define EXPECT_SUSPEND 266
#define REQUIRE_VERSION 267
#endif
#define LL_MAXTOKNO 267
#define LL_MISSINGEOF (-1)
#define LL_DELETE (0)
#define LL_VERSION 0x000505L
#define LL_NEW_TOKEN (-2)
void parse(void);
const char *LLgetSymbol(int);
extern int LLsymb;
#endif
